@@ -1,4 +1,4 @@
-// Copyright 2011 Google Inc. All rights reserved.
+// Copyright 2011 Google LLC. All rights reserved.
 // Use of this source code is governed by a BSD-style
 // license that can be found in the LICENSE file.
 
@@ -27,17 +27,17 @@ import (
 	"unicode"
 
 	"google.golang.org/api/google-api-go-generator/internal/disco"
+	"google.golang.org/api/internal/version"
 )
 
 const (
 	googleDiscoveryURL = "https://www.googleapis.com/discovery/v1/apis"
-	generatorVersion   = "2018018"
 )
 
 var (
 	apiToGenerate = flag.String("api", "*", "The API ID to generate, like 'tasks:v1'. A value of '*' means all.")
 	useCache      = flag.Bool("cache", true, "Use cache of discovered Google API discovery documents.")
-	genDir        = flag.String("gendir", "", "Directory to use to write out generated Go files")
+	genDir        = flag.String("gendir", defaultGenDir(), "Directory to use to write out generated Go files")
 	build         = flag.Bool("build", false, "Compile generated packages.")
 	install       = flag.Bool("install", false, "Install generated packages.")
 	apisURL       = flag.String("discoveryurl", googleDiscoveryURL, "URL to root discovery document")
@@ -50,12 +50,22 @@ var (
 	baseURL        = flag.String("base_url", "", "(optional) Override the default service API URL. If empty, the service's root URL will be used.")
 	headerPath     = flag.String("header_path", "", "If non-empty, prepend the contents of this file to generated services.")
 
-	gensupportPkg = flag.String("gensupport_pkg", "google.golang.org/api/gensupport", "Go package path of the 'api/gensupport' support package.")
-	googleapiPkg  = flag.String("googleapi_pkg", "google.golang.org/api/googleapi", "Go package path of the 'api/googleapi' support package.")
+	gensupportPkg     = flag.String("gensupport_pkg", "google.golang.org/api/internal/gensupport", "Go package path of the 'api/internal/gensupport' support package.")
+	googleapiPkg      = flag.String("googleapi_pkg", "google.golang.org/api/googleapi", "Go package path of the 'api/googleapi' support package.")
+	optionPkg         = flag.String("option_pkg", "google.golang.org/api/option", "Go package path of the 'api/option' support package.")
+	internalOptionPkg = flag.String("internaloption_pkg", "google.golang.org/api/option/internaloption", "Go package path of the 'api/option/internaloption' support package.")
+	htransportPkg     = flag.String("htransport_pkg", "google.golang.org/api/transport/http", "Go package path of the 'api/transport/http' support package.")
+
+	copyrightYear = flag.String("copyright_year", fmt.Sprintf("%d", time.Now().Year()), "Year for copyright.")
 
 	serviceTypes = []string{"Service", "APIService"}
 )
 
+var (
+	errOldRevision = errors.New("revision pulled older than local cached revision")
+	errNoDoc       = errors.New("could not read discovery doc")
+)
+
 // API represents an API to generate, as well as its state while it's
 // generating.
 type API struct {
@@ -112,6 +122,11 @@ func (e *compileError) Error() string {
 	return fmt.Sprintf("API %s failed to compile:\n%v", e.api.ID, e.output)
 }
 
+// skipAPIGeneration is a set of APIs to not generate when generating all clients.
+var skipAPIGeneration = map[string]bool{
+	"sql:v1beta4": true,
+}
+
 func main() {
 	flag.Parse()
 
@@ -132,7 +147,10 @@ func main() {
 		matches = append(matches, api)
 		log.Printf("Generating API %s", api.ID)
 		err := api.WriteGeneratedCode()
-		if err != nil && err != errNoDoc {
+		if err == errOldRevision {
+			log.Printf("Old revision found for %s, skipping generation", api.ID)
+			continue
+		} else if err != nil && err != errNoDoc {
 			errors = append(errors, &generateError{api, err})
 			continue
 		}
@@ -178,6 +196,9 @@ func (a *API) want() bool {
 			return false
 		}
 	}
+	if skipAPIGeneration[a.ID] && *apiToGenerate == "*" {
+		return false
+	}
 	return *apiToGenerate == "*" || *apiToGenerate == a.ID
 }
 
@@ -272,6 +293,40 @@ func apiFromFile(file string) (*API, error) {
 	return a, nil
 }
 
+func checkAndUpdateSpecFile(file string, contents []byte) error {
+	// check if file exists
+	if _, err := os.Stat(file); os.IsNotExist(err) {
+		return writeFile(file, contents)
+	}
+	existing, err := ioutil.ReadFile(file)
+	if err != nil {
+		return err
+	}
+	if err := isNewerRevision(existing, contents); err != nil {
+		return err
+	}
+	return writeFile(file, contents)
+}
+
+// isNewerRevision returns nil if the contents of new has a newer revision than
+// the contents of old.
+func isNewerRevision(old []byte, new []byte) error {
+	type docRevision struct {
+		Revision string `json:"revision"`
+	}
+	var oldDoc, newDoc docRevision
+	if err := json.Unmarshal(old, &oldDoc); err != nil {
+		return err
+	}
+	if err := json.Unmarshal(new, &newDoc); err != nil {
+		return err
+	}
+	if newDoc.Revision < oldDoc.Revision {
+		return errOldRevision
+	}
+	return nil
+}
+
 func writeFile(file string, contents []byte) error {
 	// Don't write it if the contents are identical.
 	existing, err := ioutil.ReadFile(file)
@@ -365,12 +420,17 @@ func (p *namePool) Get(preferred string) string {
 }
 
 func genDirRoot() string {
-	if *genDir != "" {
-		return *genDir
+	if *genDir == "" {
+		log.Fatalf("-gendir option must be set.")
 	}
+	return *genDir
+}
+
+func defaultGenDir() string {
+	// TODO(cbro): consider using $CWD
 	paths := filepath.SplitList(os.Getenv("GOPATH"))
 	if len(paths) == 0 {
-		log.Fatalf("No GOPATH set.")
+		return ""
 	}
 	return filepath.Join(paths[0], "src", "google.golang.org", "api")
 }
@@ -397,6 +457,12 @@ func (a *API) Target() string {
 // ServiceType returns the name of the type to use for the root API struct
 // (typically "Service").
 func (a *API) ServiceType() string {
+	if a.Name == "monitoring" && a.Version == "v3" {
+		// HACK(deklerk) monitoring:v3 should always use call its overall
+		// service struct "Service", even though there is a "Service" in its
+		// schema (we re-map it to MService later).
+		return "Service"
+	}
 	switch a.Name {
 	case "appengine", "content": // retained for historical compatibility.
 		return "APIService"
@@ -472,10 +538,9 @@ func (a *API) JSONFile() string {
 	return filepath.Join(a.SourceDir(), a.Package()+"-api.json")
 }
 
-var errNoDoc = errors.New("could not read discovery doc")
-
 // WriteGeneratedCode generates code for a.
-// It returns errNoDoc if we couldn't read the discovery doc.
+// It returns errNoDoc if we couldn't read the discovery doc or errOldRevision
+// if the API spec file being pulled in is older than the local cache.
 func (a *API) WriteGeneratedCode() error {
 	genfilename := *output
 	jsonBytes := a.jsonBytes()
@@ -485,7 +550,7 @@ func (a *API) WriteGeneratedCode() error {
 		return errNoDoc
 	}
 	if genfilename == "" {
-		if err := writeFile(a.JSONFile(), jsonBytes); err != nil {
+		if err := checkAndUpdateSpecFile(a.JSONFile(), jsonBytes); err != nil {
 			return err
 		}
 		outdir := a.SourceDir()
@@ -552,29 +617,58 @@ func (a *API) GenerateCode() ([]byte, error) {
 		}
 	}
 
-	pn(fmt.Sprintf(`// Copyright %d Google Inc. All rights reserved.
+	pn(`// Copyright %s Google LLC.
 // Use of this source code is governed by a BSD-style
 // license that can be found in the LICENSE file.
 
 // Code generated file. DO NOT EDIT.
-`, time.Now().Year()))
+`, *copyrightYear)
 
 	pn("// Package %s provides access to the %s.", pkg, a.doc.Title)
-	if r := replacementPackage[pkg]; r != "" {
+	if r := replacementPackage.Get(pkg, a.Version); r != "" {
 		pn("//")
 		pn("// This package is DEPRECATED. Use package %s instead.", r)
 	}
 	docsLink = a.doc.DocumentationLink
 	if docsLink != "" {
 		pn("//")
-		pn("// See %s", docsLink)
+		pn("// For product documentation, see: %s", docsLink)
 	}
-	pn("//\n// Usage example:")
+	pn("//")
+	pn("// Creating a client")
+	pn("//")
+	pn("// Usage example:")
 	pn("//")
 	pn("//   import %q", a.Target())
 	pn("//   ...")
-	pn("//   %sService, err := %s.New(oauthHttpClient)", pkg, pkg)
-
+	pn("//   ctx := context.Background()")
+	pn("//   %sService, err := %s.NewService(ctx)", pkg, pkg)
+	pn("//")
+	pn("// In this example, Google Application Default Credentials are used for authentication.")
+	pn("//")
+	pn("// For information on how to create and obtain Application Default Credentials, see https://developers.google.com/identity/protocols/application-default-credentials.")
+	pn("//")
+	pn("// Other authentication options")
+	pn("//")
+	if len(a.doc.Auth.OAuth2Scopes) > 1 {
+		pn(`// By default, all available scopes (see "Constants") are used to authenticate. To restrict scopes, use option.WithScopes:`)
+		pn("//")
+		// NOTE: the first scope tends to be the broadest. Use the last one to demonstrate restriction.
+		pn("//   %sService, err := %s.NewService(ctx, option.WithScopes(%s.%s))", pkg, pkg, pkg, scopeIdentifier(a.doc.Auth.OAuth2Scopes[len(a.doc.Auth.OAuth2Scopes)-1]))
+		pn("//")
+	}
+	pn("// To use an API key for authentication (note: some APIs do not support API keys), use option.WithAPIKey:")
+	pn("//")
+	pn(`//   %sService, err := %s.NewService(ctx, option.WithAPIKey("AIza..."))`, pkg, pkg)
+	pn("//")
+	pn("// To use an OAuth token (e.g., a user token obtained via a three-legged OAuth flow), use option.WithTokenSource:")
+	pn("//")
+	pn("//   config := &oauth2.Config{...}")
+	pn("//   // ...")
+	pn("//   token, err := config.Exchange(ctx, ...)")
+	pn("//   %sService, err := %s.NewService(ctx, option.WithTokenSource(config.TokenSource(ctx, token)))", pkg, pkg)
+	pn("//")
+	pn("// See https://godoc.org/google.golang.org/api/option/ for details on options.")
 	pn("package %s // import %q", pkg, a.Target())
 	p("\n")
 	pn("import (")
@@ -599,6 +693,9 @@ func (a *API) GenerateCode() ([]byte, error) {
 	}{
 		{*gensupportPkg, "gensupport"},
 		{*googleapiPkg, "googleapi"},
+		{*optionPkg, "option"},
+		{*internalOptionPkg, "internaloption"},
+		{*htransportPkg, "htransport"},
 	} {
 		pn("  %s %q", imp.lname, imp.pkg)
 	}
@@ -616,6 +713,7 @@ func (a *API) GenerateCode() ([]byte, error) {
 	pn("var _ = errors.New")
 	pn("var _ = strings.Replace")
 	pn("var _ = context.Canceled")
+	pn("var _ = internaloption.WithDefaultEndpoint")
 	pn("")
 	pn("const apiId = %q", a.doc.ID)
 	pn("const apiName = %q", a.doc.Name)
@@ -631,6 +729,31 @@ func (a *API) GenerateCode() ([]byte, error) {
 	a.GetName("New")
 	a.GetName(service)
 
+	pn("// NewService creates a new %s.", service)
+	pn("func NewService(ctx context.Context, opts ...option.ClientOption) (*%s, error) {", service)
+	if len(a.doc.Auth.OAuth2Scopes) != 0 {
+		pn("scopesOption := option.WithScopes(")
+		for _, scope := range a.doc.Auth.OAuth2Scopes {
+			pn("%q,", scope.ID)
+		}
+		pn(")")
+		pn("// NOTE: prepend, so we don't override user-specified scopes.")
+		pn("opts = append([]option.ClientOption{scopesOption}, opts...)")
+	}
+	pn("opts = append(opts, internaloption.WithDefaultEndpoint(basePath))")
+	pn("client, endpoint, err := htransport.NewClient(ctx, opts...)")
+	pn("if err != nil { return nil, err }")
+	pn("s, err := New(client)")
+	pn("if err != nil { return nil, err }")
+	pn(`if endpoint != "" { s.BasePath = endpoint }`)
+	pn("return s, nil")
+	pn("}\n")
+
+	pn("// New creates a new %s. It uses the provided http.Client for requests.", service)
+	pn("//")
+	pn("// Deprecated: please use NewService instead.")
+	pn("// To provide a custom HTTP client, use option.WithHTTPClient.")
+	pn("// If you are using google.golang.org/api/googleapis/transport.APIKey, use option.WithAPIKey with NewService instead.")
 	pn("func New(client *http.Client) (*%s, error) {", service)
 	pn("if client == nil { return nil, errors.New(\"client is nil\") }")
 	pn("s := &%s{client: client, BasePath: basePath}", service)
@@ -699,16 +822,21 @@ func (a *API) generateScopeConstants() {
 			a.p("\n")
 		}
 		n++
-		ident := scopeIdentifierFromURL(scope.URL)
+		ident := scopeIdentifier(scope)
 		if scope.Description != "" {
 			a.p("%s", asComment("\t", scope.Description))
 		}
-		a.pn("\t%s = %q", ident, scope.URL)
+		a.pn("\t%s = %q", ident, scope.ID)
 	}
 	a.p(")\n\n")
 }
 
-func scopeIdentifierFromURL(urlStr string) string {
+func scopeIdentifier(s disco.Scope) string {
+	if s.ID == "openid" {
+		return "OpenIDScope"
+	}
+
+	urlStr := s.ID
 	const prefix = "https://www.googleapis.com/auth/"
 	if !strings.HasPrefix(urlStr, prefix) {
 		const https = "https://"
@@ -819,9 +947,12 @@ var pointerFields = []fieldName{
 	{api: "cloudmonitoring:v2beta2", schema: "Point", field: "DoubleValue"},
 	{api: "cloudmonitoring:v2beta2", schema: "Point", field: "Int64Value"},
 	{api: "cloudmonitoring:v2beta2", schema: "Point", field: "StringValue"},
+	{api: "compute:alpha", schema: "ExternalVpnGateway", field: "Id"},
 	{api: "compute:alpha", schema: "Scheduling", field: "AutomaticRestart"},
+	{api: "compute:beta", schema: "ExternalVpnGateway", field: "Id"},
 	{api: "compute:beta", schema: "MetadataItems", field: "Value"},
 	{api: "compute:beta", schema: "Scheduling", field: "AutomaticRestart"},
+	{api: "compute:v1", schema: "ExternalVpnGateway", field: "Id"},
 	{api: "compute:v1", schema: "MetadataItems", field: "Value"},
 	{api: "compute:v1", schema: "Scheduling", field: "AutomaticRestart"},
 	{api: "content:v2", schema: "AccountUser", field: "Admin"},
@@ -848,11 +979,19 @@ var pointerFields = []fieldName{
 	{api: "servicecontrol:v1", schema: "MetricValue", field: "DoubleValue"},
 	{api: "servicecontrol:v1", schema: "MetricValue", field: "Int64Value"},
 	{api: "servicecontrol:v1", schema: "MetricValue", field: "StringValue"},
+	{api: "sheets:v4", schema: "ExtendedValue", field: "BoolValue"},
+	{api: "sheets:v4", schema: "ExtendedValue", field: "FormulaValue"},
+	{api: "sheets:v4", schema: "ExtendedValue", field: "NumberValue"},
+	{api: "sheets:v4", schema: "ExtendedValue", field: "StringValue"},
+	{api: "slides:v1", schema: "Range", field: "EndIndex"},
+	{api: "slides:v1", schema: "Range", field: "StartIndex"},
 	{api: "sqladmin:v1beta4", schema: "Settings", field: "StorageAutoResize"},
+	{api: "sqladmin:v1", schema: "Settings", field: "StorageAutoResize"},
 	{api: "storage:v1", schema: "BucketLifecycleRuleCondition", field: "IsLive"},
 	{api: "storage:v1beta2", schema: "BucketLifecycleRuleCondition", field: "IsLive"},
 	{api: "tasks:v1", schema: "Task", field: "Completed"},
 	{api: "youtube:v3", schema: "ChannelSectionSnippet", field: "Position"},
+	{api: "youtube:v3", schema: "MonitorStreamInfo", field: "EnableMonitorStream"},
 }
 
 // forcePointerType reports whether p should be represented as a pointer type in its parent schema struct.
@@ -1109,6 +1248,15 @@ func (s *Schema) GoName() string {
 			s.goName = s.api.typeAsGo(s.typ, false)
 		} else {
 			base := initialCap(s.apiName)
+
+			// HACK(deklerk) Re-maps monitoring's Service field to MService so
+			// that the overall struct for this API can keep its name "Service".
+			// This takes care of "Service" the initial "goName" for "Service"
+			// refs.
+			if s.api.Name == "monitoring" && base == "Service" {
+				base = "MService"
+			}
+
 			s.goName = s.api.GetName(base)
 			if base == "Service" && s.goName != "Service" {
 				// Detect the case where a resource is going to clash with the
@@ -1629,6 +1777,9 @@ func (meth *Method) generateCode() {
 	pn("\n// method id %q:", meth.Id())
 
 	retType := responseType(a, meth.m)
+	if meth.IsRawResponse() {
+		retType = "*http.Response"
+	}
 	retTypeComma := retType
 	if retTypeComma != "" {
 		retTypeComma += ", "
@@ -1844,6 +1995,7 @@ func (meth *Method) generateCode() {
 
 	pn("\nfunc (c *%s) doRequest(alt string) (*http.Response, error) {", callName)
 	pn(`reqHeaders := make(http.Header)`)
+	pn(`reqHeaders.Set("x-goog-api-client", "gl-go/"+gensupport.GoVersion()+" gdcl/%s")`, version.Repo)
 	pn("for k, v := range c.header_ {")
 	pn(" reqHeaders[k] = v")
 	pn("}")
@@ -1854,31 +2006,33 @@ func (meth *Method) generateCode() {
 		pn("}")
 	}
 	pn("var body io.Reader = nil")
-	if ba := args.bodyArg(); ba != nil && httpMethod != "GET" {
-		if meth.m.ID == "ml.projects.predict" {
-			// Skip JSONReader for APIs that require clients to pass in JSON already.
-			pn("body = strings.NewReader(c.%s.HttpBody.Data)", ba.goname)
-		} else {
-			style := "WithoutDataWrapper"
-			if a.needsDataWrapper() {
-				style = "WithDataWrapper"
+	if meth.IsRawRequest() {
+		pn("body = c.body_")
+	} else {
+		if ba := args.bodyArg(); ba != nil && httpMethod != "GET" {
+			if meth.m.ID == "ml.projects.predict" {
+				// TODO(cbro): move ML API to rawHTTP (it will be a breaking change)
+				// Skip JSONReader for APIs that require clients to pass in JSON already.
+				pn("body = strings.NewReader(c.%s.HttpBody.Data)", ba.goname)
+			} else {
+				style := "WithoutDataWrapper"
+				if a.needsDataWrapper() {
+					style = "WithDataWrapper"
+				}
+				pn("body, err := googleapi.%s.JSONReader(c.%s)", style, ba.goname)
+				pn("if err != nil { return nil, err }")
 			}
-			pn("body, err := googleapi.%s.JSONReader(c.%s)", style, ba.goname)
-			pn("if err != nil { return nil, err }")
-		}
 
-		pn(`reqHeaders.Set("Content-Type", "application/json")`)
+			pn(`reqHeaders.Set("Content-Type", "application/json")`)
+		}
+		pn(`c.urlParams_.Set("alt", alt)`)
+		pn(`c.urlParams_.Set("prettyPrint", "false")`)
 	}
-	pn(`c.urlParams_.Set("alt", alt)`)
-	pn(`c.urlParams_.Set("prettyPrint", "false")`)
 
 	pn("urls := googleapi.ResolveRelative(c.s.BasePath, %q)", meth.m.Path)
 	if meth.supportsMediaUpload() {
 		pn("if c.mediaInfo_ != nil {")
-		// Hack guess, since we get a 404 otherwise:
-		//pn("urls = googleapi.ResolveRelative(%q, %q)", a.apiBaseURL(), meth.mediaUploadPath())
-		// Further hack.  Discovery doc is wrong?
-		pn("  urls = strings.Replace(urls, %q, %q, 1)", "https://www.googleapis.com/", "https://www.googleapis.com/upload/")
+		pn("  urls = googleapi.ResolveRelative(c.s.BasePath, %q)", meth.mediaUploadPath())
 		pn(`  c.urlParams_.Set("uploadType", c.mediaInfo_.UploadType())`)
 		pn("}")
 
@@ -1889,12 +2043,12 @@ func (meth *Method) generateCode() {
 		pn("body, getBody, cleanup := c.mediaInfo_.UploadRequest(reqHeaders, body)")
 		pn("defer cleanup()")
 	}
-	pn("urls += \"?\" + c.urlParams_.Encode()")
+	pn(`urls += "?" + c.urlParams_.Encode()`)
 	pn("req, err := http.NewRequest(%q, urls, body)", httpMethod)
 	pn("if err != nil { return nil, err }")
 	pn("req.Header = reqHeaders")
 	if meth.supportsMediaUpload() {
-		pn("gensupport.SetGetBody(req, getBody)")
+		pn("req.GetBody = getBody")
 	}
 
 	// Replace param values after NewRequest to avoid reencoding them.
@@ -1907,8 +2061,11 @@ func (meth *Method) generateCode() {
 		}
 		pn(`})`)
 	}
-
-	pn("return gensupport.SendRequest(c.ctx_, c.s.client, req)")
+	if meth.supportsMediaUpload() && meth.api.Name == "storage" {
+		pn("return gensupport.SendRequestWithRetry(c.ctx_, c.s.client, req)")
+	} else {
+		pn("return gensupport.SendRequest(c.ctx_, c.s.client, req)")
+	}
 	pn("}")
 
 	if meth.supportsMediaDownload() {
@@ -1929,7 +2086,7 @@ func (meth *Method) generateCode() {
 
 	mapRetType := strings.HasPrefix(retTypeComma, "map[")
 	pn("\n// Do executes the %q call.", meth.m.ID)
-	if retTypeComma != "" && !mapRetType {
+	if retTypeComma != "" && !mapRetType && !meth.IsRawResponse() {
 		commentFmtStr := "Exactly one of %v or error will be non-nil. " +
 			"Any non-2xx status code is an error. " +
 			"Response headers are in either %v.ServerResponse.Header " +
@@ -1945,67 +2102,71 @@ func (meth *Method) generateCode() {
 		nilRet = "nil, "
 	}
 	pn(`gensupport.SetOptions(c.urlParams_, opts...)`)
-	pn(`res, err := c.doRequest("json")`)
-
-	if retTypeComma != "" && !mapRetType {
-		pn("if res != nil && res.StatusCode == http.StatusNotModified {")
-		pn(" if res.Body != nil { res.Body.Close() }")
-		pn(" return nil, &googleapi.Error{")
-		pn("  Code: res.StatusCode,")
-		pn("  Header: res.Header,")
-		pn(" }")
-		pn("}")
-	}
-	pn("if err != nil { return %serr }", nilRet)
-	pn("defer googleapi.CloseBody(res)")
-	pn("if err := googleapi.CheckResponse(res); err != nil { return %serr }", nilRet)
-	if meth.supportsMediaUpload() {
-		pn(`rx := c.mediaInfo_.ResumableUpload(res.Header.Get("Location"))`)
-		pn("if rx != nil {")
-		pn(" rx.Client = c.s.client")
-		pn(" rx.UserAgent = c.s.userAgent()")
-		pn(" ctx := c.ctx_")
-		pn(" if ctx == nil {")
-		// TODO(mcgreevy): Require context when calling Media, or Do.
-		pn("  ctx = context.TODO()")
-		pn(" }")
-		pn(" res, err = rx.Upload(ctx)")
-		pn(" if err != nil { return %serr }", nilRet)
-		pn(" defer res.Body.Close()")
-		pn(" if err := googleapi.CheckResponse(res); err != nil { return %serr }", nilRet)
-		pn("}")
-	}
-	if retTypeComma == "" {
-		pn("return nil")
+	if meth.IsRawResponse() {
+		pn(`return c.doRequest("")`)
 	} else {
-		if mapRetType {
-			pn("var ret %s", responseType(a, meth.m))
-		} else {
-			pn("ret := &%s{", responseTypeLiteral(a, meth.m))
-			pn(" ServerResponse: googleapi.ServerResponse{")
+		pn(`res, err := c.doRequest("json")`)
+
+		if retTypeComma != "" && !mapRetType {
+			pn("if res != nil && res.StatusCode == http.StatusNotModified {")
+			pn(" if res.Body != nil { res.Body.Close() }")
+			pn(" return nil, &googleapi.Error{")
+			pn("  Code: res.StatusCode,")
 			pn("  Header: res.Header,")
-			pn("  HTTPStatusCode: res.StatusCode,")
-			pn(" },")
+			pn(" }")
 			pn("}")
 		}
-		if a.needsDataWrapper() {
-			pn("target := &struct {")
-			pn("  Data %s `json:\"data\"`", responseType(a, meth.m))
-			pn("}{ret}")
-		} else {
-			pn("target := &ret")
+		pn("if err != nil { return %serr }", nilRet)
+		pn("defer googleapi.CloseBody(res)")
+		pn("if err := googleapi.CheckResponse(res); err != nil { return %serr }", nilRet)
+		if meth.supportsMediaUpload() {
+			pn(`rx := c.mediaInfo_.ResumableUpload(res.Header.Get("Location"))`)
+			pn("if rx != nil {")
+			pn(" rx.Client = c.s.client")
+			pn(" rx.UserAgent = c.s.userAgent()")
+			pn(" ctx := c.ctx_")
+			pn(" if ctx == nil {")
+			// TODO(mcgreevy): Require context when calling Media, or Do.
+			pn("  ctx = context.TODO()")
+			pn(" }")
+			pn(" res, err = rx.Upload(ctx)")
+			pn(" if err != nil { return %serr }", nilRet)
+			pn(" defer res.Body.Close()")
+			pn(" if err := googleapi.CheckResponse(res); err != nil { return %serr }", nilRet)
+			pn("}")
 		}
+		if retTypeComma == "" {
+			pn("return nil")
+		} else {
+			if mapRetType {
+				pn("var ret %s", responseType(a, meth.m))
+			} else {
+				pn("ret := &%s{", responseTypeLiteral(a, meth.m))
+				pn(" ServerResponse: googleapi.ServerResponse{")
+				pn("  Header: res.Header,")
+				pn("  HTTPStatusCode: res.StatusCode,")
+				pn(" },")
+				pn("}")
+			}
+			if a.needsDataWrapper() {
+				pn("target := &struct {")
+				pn("  Data %s `json:\"data\"`", responseType(a, meth.m))
+				pn("}{ret}")
+			} else {
+				pn("target := &ret")
+			}
 
-		if meth.m.ID == "ml.projects.predict" {
-			pn("var b bytes.Buffer")
-			pn("if _, err := io.Copy(&b, res.Body); err != nil { return nil, err }")
-			pn("if err := res.Body.Close(); err != nil { return nil, err }")
-			pn("if err := json.NewDecoder(bytes.NewReader(b.Bytes())).Decode(target); err != nil { return nil, err }")
-			pn("ret.Data = b.String()")
-		} else {
-			pn("if err := gensupport.DecodeResponse(target, res); err != nil { return nil, err }")
+			if meth.m.ID == "ml.projects.predict" {
+				pn("var b bytes.Buffer")
+				pn("if _, err := io.Copy(&b, res.Body); err != nil { return nil, err }")
+				pn("if err := res.Body.Close(); err != nil { return nil, err }")
+				pn("if err := json.NewDecoder(bytes.NewReader(b.Bytes())).Decode(target); err != nil { return nil, err }")
+				pn("ret.Data = b.String()")
+			} else {
+				pn("if err := gensupport.DecodeResponse(target, res); err != nil { return nil, err }")
+			}
+			pn("return ret, nil")
 		}
-		pn("return ret, nil")
 	}
 
 	bs, err := json.MarshalIndent(meth.m.JSONMap, "\t// ", "  ")
@@ -2116,6 +2277,32 @@ func resolveRelative(basestr, relstr string) string {
 	return u.String()
 }
 
+func (meth *Method) IsRawRequest() bool {
+	if meth.m.Request == nil {
+		return false
+	}
+	// TODO(cbro): enable across other APIs.
+	if meth.api.Name != "healthcare" {
+		return false
+	}
+	return meth.m.Request.Ref == "HttpBody"
+}
+
+func (meth *Method) IsRawResponse() bool {
+	if meth.m.Response == nil {
+		return false
+	}
+	if meth.IsRawRequest() {
+		// always match raw requests with raw responses.
+		return true
+	}
+	// TODO(cbro): enable across other APIs.
+	if meth.api.Name != "healthcare" {
+		return false
+	}
+	return meth.m.Response.Ref == "HttpBody"
+}
+
 func (meth *Method) NewArguments() *arguments {
 	args := &arguments{
 		method: meth,
@@ -2134,7 +2321,14 @@ func (meth *Method) NewArguments() *arguments {
 		args.AddArg(arg)
 	}
 	if rs := meth.m.Request; rs != nil {
-		args.AddArg(meth.NewBodyArg(rs))
+		if meth.IsRawRequest() {
+			args.AddArg(&argument{
+				goname: "body_",
+				gotype: "io.Reader",
+			})
+		} else {
+			args.AddArg(meth.NewBodyArg(rs))
+		}
 	}
 	return args
 }
